/*
 * Written by Marco d'Itri <md@linux.it>.
 * Many thanks to the anonymous who provided find_sc().
 *
 * You can show your appreciation for this at http://blog.bofh.it/amazon :-)
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/unistd.h>
#include <linux/sched.h>

#ifndef __i386__
#error Unsupported architecture
#endif

#ifndef __NR_vmsplice
#error This system does not support vmsplice(2)
#endif


static void **sys_call_table = NULL;

static long (*old_vmsplice)(int, const struct iovec *, unsigned long, unsigned int);
#ifdef __i386__
static long (*old_vm86old)(struct vm86_struct *);
#endif


static void find_sc(void)
{
	int i;
	unsigned long *ptr;
	unsigned long arr[4];
	extern int sys_close(int fd);

	ptr = (unsigned long *)((init_mm.end_code + 4) & 0xfffffffc);

	while ((unsigned long)ptr < (unsigned long)init_mm.end_data) {
		if (*ptr == (unsigned long)((unsigned long *)sys_close)) {
			for (i = 0; i < 4; i++) {
				arr[i] = *(ptr + i);
				arr[i] = (arr[i] >> 16) & 0x0000ffff;
			}

			if (arr[0] != arr[2] || arr[1] != arr[3]) {
				sys_call_table = (ptr - __NR_close);
				break;
			}
		}
		ptr++;
	}
}

static long nosuchsyscall(int fd, const struct iovec *iov,
			   unsigned long nr_segs, unsigned int flags)
{
#ifndef BE_QUIET
	BUG();
#endif
	return -EINVAL;
}

int init_module(void)
{
	find_sc();
	if (!sys_call_table)
	    return -EIO;
	printk(__FILE__ ": sys_call_table is at %p\n", sys_call_table);

	old_vmsplice = sys_call_table[__NR_vmsplice];
	sys_call_table[__NR_vmsplice] = nosuchsyscall;
#ifdef __i386__
	old_vm86old = sys_call_table[__NR_vm86old];
	sys_call_table[__NR_vm86old] = nosuchsyscall;
#endif
	return 0;
}

void cleanup_module(void)
{
	sys_call_table[__NR_vmsplice] = old_vmsplice;
#ifdef __i386__
	sys_call_table[__NR_vm86old] = old_vm86old;
#endif
}

MODULE_LICENSE("GPL");
